// SPDX-License-Identifier: GPL-2.0-or-later
/** @file
 * TODO: insert short description here
 *//*
 * Authors: see git history
 *
 * Copyright (C) 2018 Authors
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */
#ifndef CONDITIONS_H_
#define CONDITIONS_H_

class SPItem;

bool sp_item_evaluate(SPItem const *item);

#endif /*CONDITIONS_H_*/
