// SPDX-License-Identifier: GPL-2.0-or-later
#ifndef SEEN_TOOLBAR_CONSTANTS_H
#define SEEN_TOOLBAR_CONSTANTS_H

/*
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */

namespace Inkscape::UI::Toolbar {

static constexpr const char* tools_icon_size = "/toolbox/tools/iconsize";
static constexpr const char* ctrlbars_icon_size = "/toolbox/controlbars/iconsize";
static constexpr const int min_pixel_size = 16;
static constexpr const int max_pixel_size = 48;

} // namespace Inkscape::UI::Toolbar

#endif /* !SEEN_TOOLBAR_CONSTANTS_H */

/*
  Local Variables:
  mode:c++
  c-file-style:"stroustrup"
  c-file-offsets:((innamespace . 0)(inline-open . 0)(case-label . +))
  indent-tabs-mode:nil
  fill-column:99
  End:
*/
// vim: filetype=cpp:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:fileencoding=utf-8:textwidth=99 :
