// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Gio::Actions tied to the application and without GUI.
 *
 * Copyright (C) 2018 Tavmjong Bah
 *
 * The contents of this file may be used under the GNU General Public License Version 2 or later.
 *
 */

#ifndef INK_ACTIONS_BASE_H
#define INK_ACTIONS_BASE_H

class InkscapeApplication;

void add_actions_base(InkscapeApplication* app);

#endif // INK_ACTIONS_BASE_H
