// SPDX-License-Identifier: GPL-2.0-or-later

#include "inkscape-version.h"

namespace Inkscape {
    char const *revision_string = "dcaf3e7d9e";
    char const *version_string = "1.4.4" " " "(dcaf3e7d9e, 2026-05-05)";
    char const *version_string_without_revision = "1.4.4";

    unsigned short int const build_year = 2026;

    unsigned int const version_major = 1;
    unsigned int const version_minor = 4;
    unsigned int const version_patch = 4;
}
