// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * Provide customizable shortcut keys to tools
 *
 * Authors:
 *   Martin Owens
 *
 * Copyright (C) 2024 Authors
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */

#ifndef UI_TOOLS_SHORTCUTS
#define UI_TOOLS_SHORTCUTS

#include <string>

class InkscapeApplication;

void init_tool_shortcuts(InkscapeApplication* app);

#endif // UI_TOOLS_SHORTCUTS

/*
  Local Variables:
  mode:c++
  c-file-style:"stroustrup"
  c-file-offsets:((innamespace . 0)(inline-open . 0)(case-label . +))
  indent-tabs-mode:nil
  fill-column:99
  End:
*/
// vim: filetype=cpp:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:fileencoding=utf-8:textwidth=99 :
