// SPDX-License-Identifier: GPL-2.0-or-later
/** \file
 *
 * Authors:
 *   Sushant A A <sushant.co19@gmail.com>
 *
 * Copyright (C) 2021 Authors
 *
 * Released under GNU GPL v2+, read the file 'COPYING' for more information.
 */

#ifndef INK_ACTIONS_TEXT_H
#define INK_ACTIONS_TEXT_H

class InkscapeApplication;

void add_actions_text(InkscapeApplication* app);

#endif // INK_ACTIONS_TEXT_H
